@@ -49,7 +49,25 @@ def cmake_exists():
     path = find_executable('cmake3')
     return path is not None, path 
 
+def is_rhel7():
+    """check if we are running on rhel7 platform"""
+    try:
+      f = open( '/etc/redhat-release', "r" )
+      txt = f.read().split()
+      i = txt.index( 'release' ) + 1
+      return txt[i][0] == '7'
+    except IOError:
+      return False
+    except ValueError:
+      return False
 
+def has_devtoolset():
+    """check if devtoolset-7 is installed"""
+    import subprocess
+    args = ( "/usr/bin/rpm", "-q", "devtoolset-7" )
+    popen = subprocess.Popen(args, stdout=subprocess.PIPE)
+    rc = popen.wait()
+    return rc == 0
 
 # def python_dependency_name( py_version_short, py_version_nodot ):
 #     """ find the name of python dependency """
@@ -77,9 +95,15 @@ class CustomInstall(install):
         comp_bin    = binary_exists( 'c++' ) or binary_exists( 'g++' ) or binary_exists( 'clang' )
 
         import pkgconfig
-        zlib_dev    = pkgconfig.exists( 'zlib' )
-        openssl_dev = pkgconfig.exists( 'openssl' )
-        uuid_dev    = pkgconfig.exists( 'uuid' )
+        zlib_dev     = pkgconfig.exists( 'zlib' )
+        openssl_dev  = pkgconfig.exists( 'openssl' )
+        uuid_dev     = pkgconfig.exists( 'uuid' )
+        if is_rhel7():
+          devtoolset7 = has_devtoolset()
+          need_devtoolset = "true"
+        else:
+          devtoolset7 = True # we only care about devtoolset7 on rhel7
+          need_devtoolset = "false"
         
         pyname = None
         if py_version_nodot[0] == '3':
@@ -89,17 +113,18 @@ class CustomInstall(install):
             python_dev = pkgconfig.exists( 'python' );
             pyname = 'python'
 
-        missing_dep = not ( cmake_bin and make_bin and comp_bin and zlib_dev and openssl_dev and python_dev and uuid_dev )
+        missing_dep = not ( cmake_bin and make_bin and comp_bin and zlib_dev and openssl_dev and python_dev and uuid_dev and devtoolset7 )
 
         if missing_dep:
           print( 'Some dependencies are missing:')
-          if not cmake_bin:   print('\tcmake (version 3) is missing!')
-          if not make_bin:    print('\tmake is missing!')
-          if not comp_bin:    print('\tC++ compiler is missing (g++, c++, clang, etc.)!')
-          if not zlib_dev:    print('\tzlib development package is missing!')
-          if not openssl_dev: print('\topenssl development package is missing!')
-          if not python_dev:  print('\t{} development package is missing!'.format(pyname) )
-          if not uuid_dev:    print('\tuuid development package is missing')
+          if not cmake_bin:    print('\tcmake (version 3) is missing!')
+          if not make_bin:     print('\tmake is missing!')
+          if not comp_bin:     print('\tC++ compiler is missing (g++, c++, clang, etc.)!')
+          if not zlib_dev:     print('\tzlib development package is missing!')
+          if not openssl_dev:  print('\topenssl development package is missing!')
+          if not python_dev:   print('\t{} development package is missing!'.format(pyname) )
+          if not uuid_dev:     print('\tuuid development package is missing')
+          if not devtoolset7:  print('\tdevtoolset-7 package is missing')
           raise Exception( 'Dependencies missing!' )
 
         useropt = ''
@@ -113,6 +138,8 @@ class CustomInstall(install):
         command.append( py_version_short )
         command.append( useropt )
         command.append( cmake_path )
+        command.append( need_devtoolset )
+        command.append( sys.executable )
         rc = subprocess.call(command)
         if rc:
           raise Exception( 'Install step failed!' )
@@ -135,11 +162,7 @@ class CustomWheelGen(bdist_wheel):
     def run(self):
         pass
 
-
-version = get_version()
-if version.startswith('unknown'):
-    version = get_version_from_file()
-
+version = get_version_from_file()
 setup_requires=[ 'pkgconfig' ]
 
 setup( 
