@@ -38,10 +38,12 @@
 #include "XrdSys/XrdSysError.hh"
 #include "XrdSys/XrdSysPthread.hh"
 #include "XrdSfs/XrdSfsInterface.hh"
+#include "XrdXrootd/XrdXrootdAioFob.hh"
 #include "XrdXrootd/XrdXrootdFile.hh"
 #include "XrdXrootd/XrdXrootdFileLock.hh"
 #include "XrdXrootd/XrdXrootdMonFile.hh"
 #include "XrdXrootd/XrdXrootdMonitor.hh"
+#include "XrdXrootd/XrdXrootdPgwFob.hh"
 #define  TRACELINK this
 #include "XrdXrootd/XrdXrootdTrace.hh"
  
@@ -81,8 +83,9 @@ static const unsigned long  heldMask = ~1UL;
 XrdXrootdFile::XrdXrootdFile(const char *id, const char *path, XrdSfsFile *fp,
                              char mode, bool async, struct stat *sP)
                             : XrdSfsp(fp), mmAddr(0), FileKey(strdup(path)),
-                              FileMode(mode), AsyncMode(async), fhProc(0),
-                              ID(id)
+                              FileMode(mode), AsyncMode(async),
+                              aioFob(0), pgwFob(0), fhProc(0),
+                              ID(id), refCount(0), syncWait(0)
 {
     static XrdSysMutex seqMutex;
     struct stat buf;
@@ -119,11 +122,10 @@ XrdXrootdFile::XrdXrootdFile(const char *id, const char *path, XrdSfsFile *fp,
 }
   
 /******************************************************************************/
-/*                                                                            */
 /*                                  I n i t                                   */
 /******************************************************************************/
   
-void XrdXrootdFile::Init(XrdXrootdFileLock *lp, XrdSysError *erP, int sfok)
+void XrdXrootdFile::Init(XrdXrootdFileLock *lp, XrdSysError *erP, bool sfok)
 {
    Locker = lp;
    eDest  = erP;
@@ -136,6 +138,9 @@ void XrdXrootdFile::Init(XrdXrootdFileLock *lp, XrdSysError *erP, int sfok)
   
 XrdXrootdFile::~XrdXrootdFile()
 {
+   if (aioFob) aioFob->Reset();
+
+   Serialize(); // Make sure there are no outstanding references
 
    if (XrdSfsp)
       {TRACEI(FS, "closing " <<FileMode <<' ' <<FileKey);
@@ -146,9 +151,48 @@ XrdXrootdFile::~XrdXrootdFile()
 
    if (fhProc) fhProc->Avail(fHandle);
 
-   if (FileKey) free(FileKey);
+   if (aioFob) delete aioFob;
+
+   if (pgwFob) delete pgwFob;
+
+   if (FileKey) free(FileKey); // Must be the last thing deleted!
 }
 
+/******************************************************************************/
+/*                                   R e f                                    */
+/******************************************************************************/
+  
+void XrdXrootdFile::Ref(int num)
+{
+
+// Change the reference counter and check if anyone is waiting
+//
+   fileMutex.Lock();
+   refCount += num;
+   TRACEI(DEBUG,"File::Ref "<<refCount<<" after +"<<num);
+   if (num < 0 && syncWait && refCount <= 0) syncWait->Post();
+   fileMutex.UnLock();
+}
+
+/******************************************************************************/
+/*                             S e r i a l i z e                              */
+/******************************************************************************/
+  
+void XrdXrootdFile::Serialize()
+{
+
+// Wait until the reference count reaches zero
+//
+   fileMutex.Lock();
+   if (refCount > 0)
+      {XrdSysSemaphore mySem(0);
+       TRACEI(DEBUG, "serializing access " <<FileMode <<' ' <<FileKey);
+       syncWait = &mySem;
+       fileMutex.UnLock();
+       mySem.Wait();
+      } else fileMutex.UnLock();
+}
+  
 /******************************************************************************/
 /*                   x r d _ F i l e T a b l e   C l a s s                    */
 /******************************************************************************/
@@ -251,6 +295,13 @@ XrdXrootdFile *XrdXrootdFileTable::Del(XrdXrootdMonitor *monP, int fnum,
 
    if (fp)
       {XrdXrootdFileStats &Stats = fp->Stats;
+   //!!! For now we add pgreads to normal reads and pgwrite to normal writes
+   //!!! Once we figure out how to report them separately, we need to do this.
+
+       Stats.xfr.read  += Stats.prw.rBytes;
+       Stats.xfr.write += Stats.prw.wBytes;
+       Stats.ops.read  += Stats.prw.rCount;
+       Stats.ops.write += Stats.prw.wCount; // Doesn't include retries!!!
 
        if (monP) monP->Close(Stats.FileID,
                              Stats.xfr.read + Stats.xfr.readv,
